from .constants import BATCH_DIMS, MILD_TOLERANCES, STRICT_TOLERANCES, TOLERANCES
from .equivariance import check_pin_equivariance, check_pin_invariance
from .geometric_algebra import (
    check_consistence_with_geometric_product,
    check_consistence_with_grade_involution,
    check_consistence_with_reversal,
)
